#include <stdio.h>


#define MAIN_PROGRAM
#include "globals.h"
#include "chinput2.h"
#include "pydb.h"
#include "read_conf.h"
#include "util.h"
#include "xim.h"


#define RESET_STR_PINYIN do{ \
    str_pinyin[0]='\0';      \
    str_pinyin_pos=0;        \
}while(0)

#define UPDATE_STR_PINYIN do{    \
    if( last_selected_pinyin_pos){  \
       (void)memmove(str_pinyin,(u_char *)&str_pinyin[last_selected_pinyin_pos],str_pinyin_pos-last_selected_pinyin_pos);      \
       str_pinyin_pos-=last_selected_pinyin_pos;                                                                               \
    }                                                                                                                          \
    str_pinyin[str_pinyin_pos]='\0';                                                                                           \
}while(0)

#define RESET_PHRASE_NUMBER do{		        \
    phrase_num=0;				\
    phrase_top_num=0;				\
    phrase_base_num=0;				\
    phrase_num_offset=0;			\
}while(0)

#define  HEAD_LEN     6
static XChar2b         ch2b_head[2][2][HEAD_LEN+1] = {
  { /* jian ti & update */
    {{0xa1, 0xbc}, {0xbc, 0xf2}, {0xcc, 0xe5}, {0xc8,0xab},{0xc6,0xb4},{0xa1, 0xbd}}, 
    /* jian ti & NO update */
    {{0xa1, 0xbe}, {0xbc, 0xf2}, {0xcc, 0xe5}, {0xc8,0xab},{0xc6,0xb4},{0xa1, 0xbf}}  
  },
  { /* fan  ti & update */
    {{0xa1, 0xbc}, {0xb7, 0xb1}, {0xcc, 0xe5}, {0xc8,0xab},{0xc6,0xb4},{0xa1, 0xbd}}, 
    /* fan  ti & NO update */
    {{0xa1, 0xbe}, {0xb7, 0xb1}, {0xcc, 0xe5}, {0xc8,0xab},{0xc6,0xb4},{0xa1, 0xbf}}  
  }
};


void
get_gc(Window w, GC * gc, XFontStruct * font_info)
{
	unsigned long   valuemask = 0;
	XGCValues       values;

	*gc = XCreateGC(display, w, valuemask, &values);
	XSetFont(display, *gc, font_info->fid);
	XSetForeground(display, *gc, BlackPixel(display, screen));
	XSetBackground(display, *gc, WhitePixel(display, screen));
	XSetFillStyle(display, *gc, FillSolid);
}

int
load_font(XFontStruct ** font_info, char *font_name)
{
	if ((*font_info = XLoadQueryFont(display, font_name)) == NULL)
	  return -1;
	else 
	  return 0;
}

void
load_all_need_font(void)
{
	font_type = GL_TYPE;
	if( load_font(&gb_font, FONT_GB) < 0)
	  if( load_font(&gb_font,FONT_GB_SUN) < 0)
	    err(-1,"Can not load chinese font %s and %s",
		FONT_GB,FONT_GB_SUN);
	(void)load_font(&roman_font, FONT_ROMAN);
}


void
get_font_info(XFontStruct * font_struct, XCharStruct * font_info)
{
	XChar2b         ch;
	int             direction, font_ascent, font_descent;

	/* ַ "" */
	ch.byte1 = 0x30;
	ch.byte2 = 0x21;
	XTextExtents16(font_struct, &ch, 1, &direction, &font_ascent,
		       &font_descent, font_info);
#if 0
	printf("%d\t%d\t%d\n
	       % d \ t % d \ t % d \ n
	       % d \ t % d \ t % d \ n ",
	       direction, font_ascent, font_descent,
	       font_info->lbearing,
	       font_info->rbearing,
	       font_info->width,
	       font_info->ascent,
	       font_info->descent,
	       font_info->attributes);
#endif
}


void
set_size_hints(XSizeHints * s)
{
  	s->flags = USPosition | PSize | PMinSize | PMaxSize; 
	s->x = WIN_DEFAULT_X;
	s->y = WIN_DEFAULT_Y;
	s->max_width = s->min_width = s->width = WIN_MIN_WIDTH;
	s->max_height = s->min_height = s->height = WIN_MIN_HEIGHT;
}


void
draw_head(void)
{
	adjust_code_type((u_char *) ch2b_head[encoding][need_update]);
	XDrawImageString16(display, w_input, gb_gc, 
			   WIN_HEAD_X, WIN_HEAD_Y, 
			   ch2b_head[encoding][need_update], HEAD_LEN);
}

void
clear_pinyin_area(void)
{
 XDrawImageString(display, w_input, roman_gc, WIN_PY_X, WIN_PY_Y, 
		  str_pinyin_area,
		  MAX_PINYIN_LEN);
}

void
draw_py(void)
{
    clear_pinyin_area();
  if (str_pinyin){
#if 0
    printf("str_pinyin=[%s],str_pinyin_pos=%d\n", str_pinyin,str_pinyin_pos);
#endif
    XDrawImageString(display, w_input, roman_gc, WIN_PY_X, WIN_PY_Y, 
		     str_pinyin,
		     str_pinyin_pos );
  }
}

void
clear_select_area(void)
{
	XDrawImageString(display, w_input, roman_gc, WIN_SELECT_X, WIN_SELECT_Y,
			 select_area,
			 WIN_MIN_WIDTH / roman_font_info.width+1);
}


void
draw_select(void)
{
	clear_select_area();
	assemble_select_area();
}

void
redraw(void)
{  
        XRaiseWindow(display, w_input);
	draw_head();
	draw_py();
	draw_select();
}

/*
 * Need complete in the future.
 */
void
exit_chinput2(int ret)
{
  exit(0);
}

int
process_keys(XKeyEvent * event)
{
  int             ret,action=NEED_FORWARD;
  unsigned char   buf[MAX_PINYIN_LEN];
  KeySym          key_sym;
  XComposeStatus  cs;

  ret = XLookupString((XKeyEvent *) event, (char *) buf,
		      sizeof(buf) - 1, &key_sym, &cs);
  if( event->state & ControlMask){
    switch(key_sym){
    case XK_F1:
      encoding=!encoding;
      draw_head();
      break;
    case XK_F4:
      exit_chinput2(0);
    default:
      ;
    }
  }
  buf[ret] = 0;
#if 0
  printf("key=[%s],ret=%d\n",buf,ret);
#endif
  if (!ret){
    return NEED_FORWARD;
  }else{
    if( event->state & SPECIAL_KEY_MASK ){
      return NEED_FORWARD;
    }else{
      if(event->state & ControlMask){
	switch(key_sym){
	case XK_period:
	  RESET_STR_PINYIN;
	  RESET_PHRASE_NUMBER;
	  redraw();
	  return NEED_BYPASS;
	case XK_comma:
	  need_update=!need_update;
	  draw_head();
	  return NEED_BYPASS;
	default:
	  return NEED_FORWARD;
	}
      }
    }
  }
  if (key_sym >= XK_a && key_sym <= XK_z){
    if (str_pinyin_pos < (MAX_PINYIN_LEN - ret)) {
      u_strcat(str_pinyin, buf);
      ++str_pinyin_pos;
      str_pinyin[str_pinyin_pos] = '\0';
      RESET_PHRASE_NUMBER;
      phrase_num=get_select();
#if 0
      printf("phrase_num=%d\n",phrase_num);
#endif		 
      redraw();
      return NEED_BYPASS;
    }else{
      XBell(display,BELL_VOLUMN);
      return NEED_BYPASS;
    }
  }

  switch (key_sym) {
  case XK_space:
    if(phrase_num){
      u_strcpy(commit_str,phrase[phrase_base_num]);
      update_pydb();
      UPDATE_STR_PINYIN;
      RESET_PHRASE_NUMBER;
      phrase_num=get_select();
      action=NEED_COMMIT;
    }else{
      if(str_pinyin_pos){
	XBell(display,BELL_VOLUMN);
	action=NEED_BYPASS;
      }else{
	action=NEED_FORWARD;
      }
    }
    break;
  case XK_BackSpace:
  case XK_Delete:
    if (str_pinyin_pos) {
      int last_valid_position=0;
      int i;

      (void)is_last_pinyin_valid(str_pinyin,NULL,&last_valid_position);
      for(i=last_valid_position;i<str_pinyin_pos;++i)
	str_pinyin[i]='\0';
      str_pinyin_pos=last_valid_position;
#if 0
      printf("last_valid_position=%d\n",last_valid_position);
#endif

      /*    str_pinyin[--str_pinyin_pos] = '\0';  */
      RESET_PHRASE_NUMBER;
      phrase_num=get_select(); 
      action=NEED_BYPASS;
    } else{
      RESET_PHRASE_NUMBER;
      action=NEED_FORWARD;
    }
    break;
  case XK_equal:
    if(str_pinyin_pos ){
      if( phrase_num_offset*MAX_DISP_PHRASE_NUM <= phrase_num )
	++phrase_num_offset;
      if(phrase_num_offset*MAX_DISP_PHRASE_NUM > phrase_num)
	--phrase_num_offset;
#if 0
      printf("phrase_num_offset=%d\n",phrase_num_offset);
#endif
      action=NEED_BYPASS;
    }else{
      action= NEED_FORWARD;
    }
    break;
  case XK_minus:
    if( str_pinyin_pos){
      if(phrase_num_offset > 0)
	--phrase_num_offset;
#if 0
      printf("phrase_num_offset=%d\n",phrase_num_offset);
#endif
      action=NEED_BYPASS;
    }else{
      action=NEED_FORWARD;
    }
    break;
  case XK_0:case XK_1:case XK_2:case XK_3:case XK_4:  
  case XK_5:case XK_6:case XK_7:case XK_8:case XK_9:
    if(phrase_num){
      if( key_sym == XK_0)
	key_sym=XK_1;
      if( ((key_sym -XK_1) >= MAX_DISP_PHRASE_NUM) || ((key_sym - XK_1) >= phrase_num)){
	XBell(display,BELL_VOLUMN);
	action=NEED_BYPASS;
	break;
      }
      u_strcpy(commit_str,phrase[phrase_base_num+key_sym-XK_1]);
      update_pydb();
      UPDATE_STR_PINYIN;
      RESET_PHRASE_NUMBER;
      phrase_num=get_select();
      action=NEED_COMMIT;
    }else{
      if( str_pinyin_pos){
	XBell(display,BELL_VOLUMN);
	action=NEED_BYPASS;
      }else{
	action=NEED_FORWARD;
      }
    }
    break;
  case XK_period:
    switch(font_type){
    case GL_TYPE:
      commit_str[0]=0xa1;
      commit_str[1]=0xa3;
      commit_str[2]='\0';
      break;
    case GR_TYPE:
      commit_str[0]=0x21;
      commit_str[1]=0x23;
      commit_str[2]='\0';
      break;
    default:
      warn("no such font_type [%d]\n",font_type);
      break;
    }
    action=NEED_COMMIT;
    break;
  default:
    /* char should be forwarded to application. */
    action= NEED_FORWARD ;
  }       
  redraw();
  return action;
}

static void
real_exit(int i)
{
	if( i == SIGBUS)
	 	abort();
	exit(i);
}

void 
sig_exit(int sig)
{
  fprintf(stderr,"Received signal %d,exit.\n",sig);
  real_exit(sig);
}

static void
setsig(int sig,void (*f)(int) )
{
  struct sigaction   sa;
   
  sa.sa_handler = f;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = 0;
  sigaction(sig, &sa, NULL);
}

void
set_sigproc(void)
{
  setsig(SIGBUS,sig_exit);
  setsig(SIGINT,SIG_IGN);
  setsig(SIGHUP,SIG_IGN);
  setsig(SIGPIPE,keep_alive);
}

void
init_all(int argc, char *argv[])
{
  char *tmp;

  progname = argv[0];
 
  read_conf(DEFAULT_CONF_PATH);
  init_pydb();

  memset(select_area,SPACE, sizeof(select_area));
  memset(str_pinyin_area,SPACE, sizeof(str_pinyin_area));
  memset(str_pinyin, 0, sizeof(str_pinyin));
  memset(commit_str,0,sizeof(commit_str));

  ch2b_select_pos = 0;
  str_pinyin_pos = 0;
  phrase_num=0;
  phrase_base_num=0;
  phrase_top_num=0;
  phrase_num_offset=0;
  need_update=1;
  encoding=0;
  last_selected_pinyin_pos=0;

  tmp=getenv("LC_ALL");
  if(tmp){
    if(strstr(tmp, GB_LOCALE)  != NULL){
      encoding=GB_ENCODING;
    }
    if(strstr(tmp,BIG5_LOCALE) != NULL){
      encoding=BIG5_ENCODING;
    }
  }else{
    encoding=GB_ENCODING;
  }
  tmp=getenv("XMODIFIERS");
  if(!tmp){
    setenv("XMODIFIERS","@im=Chinput",0);
  }
  set_sigproc();
}

/*
 * When using Emacs20.7,please add the following two lines in ~/.emacs file.
 *  (set-language-environment "Chinese-gb")
 *  (set-keyboard-coding-system 'cn-gb-2312)
 * and you should set the GB fontset correctly.
 *
 * and put following lines into ~/Emacs
 * Emacs.font: 8x16
 * Emacs.Fontset-0: hanzigb16fs,\ 
 * chinese-gb: hanzigb16fs, \
 * ascii: 8x16
 */
int
main(int argc, char *argv[])
{
	XSizeHints      size_hints;
	XEvent          event;
	
	init_all(argc, argv);
	if ((display = XOpenDisplay(display_name)) == NULL)
		err(-1, "%s: connot connect to X server %s\n",
		    progname, XDisplayName(display_name));
	screen = DefaultScreen(display);
	w_input = XCreateSimpleWindow(display, RootWindow(display, screen),
				      WIN_DEFAULT_X,WIN_DEFAULT_Y,
				      WIN_MIN_WIDTH, WIN_MIN_HEIGHT, 
				      WIN_MIN_BORDER,
				      BlackPixel(display, screen),
				      WhitePixel(display, screen));
	XStoreName(display, w_input, WIN_NAME);
	set_size_hints(&size_hints);
	XSetStandardProperties(display, w_input, WIN_NAME, NULL,
			       0, argv, argc, &size_hints);
	XSelectInput(display, w_input, EVENT_MASK);
	
	load_all_need_font();
	
	get_gc(w_input, &gb_gc, gb_font);
	get_gc(w_input, &roman_gc, roman_font);

	get_font_info(gb_font, &gb_font_info);
	get_font_info(roman_font, &roman_font_info);

	set_xim();

	XMapWindow(display, w_input);

	while (1) {
		XNextEvent(display, &event);
		if (XFilterEvent(&event, None) == True)
		  continue;

		switch (event.type) {
		case MappingNotify:
			XRefreshKeyboardMapping((XMappingEvent *) & event);
			break;
		case Expose:
			redraw();
			break;
		default:
			;
		}
	}
}




