#include "chinput2.h"
#include "db_util.h"
#include "globals.h"
#include "ic.h"
#include "pydb.h"
#include "xim.h"
#include "convert.h"

static Bool     im_get_ic_value(register XIMS ims, register IMChangeICStruct * call_data);
static Bool     im_set_ic_value(register XIMS ims, register IMChangeICStruct * call_data);
static Bool     im_open(register XIMS ims, register IMOpenStruct * call_data);
static Bool     im_close(register XIMS ims, register IMCloseStruct * call_data);
static Bool     im_create_ic(register XIMS ims, register IMChangeICStruct * call_data);
static Bool     im_forward_event(register XIMS ims, register IMForwardEventStruct * call_data);
static Bool     im_trigger_notify(register XIMS ims, register IMTriggerNotifyStruct * call_data);
static Bool     im_preedit_start_reply(register XIMS ims, register IMPreeditCBStruct * call_data);
static Bool     im_preedit_caret_reply(register XIMS ims, register IMPreeditCBStruct * call_data);
static Bool     im_protocol_handler(register XIMS ims, register IMProtocol * call_data);

XIMS            ims;
XIMStyles      *input_styles;
XIMTriggerKeys *on_keys;
XIMEncodings   *encodings;

Bool            use_trigger = True;	/* Dynamic Event Flow is default */

/* Supported Inputstyles */
static XIMStyle Styles[] = {
  /*
	XIMPreeditCallbacks | XIMStatusCallbacks,
	XIMPreeditPosition | XIMStatusArea,
	XIMPreeditPosition | XIMStatusNothing,
	XIMPreeditArea | XIMStatusArea,
  */
	XIMPreeditNothing | XIMStatusNothing,
	0
};


/* Trigger Keys List */
static XIMTriggerKey Trigger_Keys[] = {
	{XK_space, ShiftMask, ShiftMask},
	{0L, 0L, 0L}
};

/* Supported Encodings */
static XIMEncoding chEncodings[] = {
    	"COMPOUND_TEXT",
    	NULL
};


Bool
im_get_ic_value(register XIMS ims, register IMChangeICStruct * call_data)
{
	GetIC(call_data);
	return True;
}

Bool
im_set_ic_value(register XIMS ims, register IMChangeICStruct * call_data)
{
	SetIC(call_data);
	return True;
}

Bool
im_open(register XIMS ims, register IMOpenStruct * call_data)
{
	return True;
}

Bool
im_close(register XIMS ims, register IMCloseStruct * call_data)
{
	return True;
}

Bool
im_create_ic(register XIMS ims, register IMChangeICStruct * call_data)
{
	CreateIC(call_data);
	return True;
}
Bool
im_destroy_ic(register XIMS ims, register IMChangeICStruct * call_data)
{
	DestroyIC(call_data);
	return True;
}


Bool
im_forward_event(register XIMS ims, register IMForwardEventStruct * call_data)
{
  int ret;
  char           *text = (char *)commit_str;
  XTextProperty   tp;
  XKeyEvent *key_event;
  static char for_big5[MAX_HANZI_CHAR_NUMBER*2+1];

  key_event=(XKeyEvent *)&call_data->event;

  /* Lookup KeyPress Events only */
  if (key_event->type != KeyPress)
    return True;

  ret=process_keys(key_event);
  switch(ret){
  case NEED_COMMIT:
    if( commit_str[0] == '\0')
      return True;
    if( encoding == BIG5_ENCODING){
      gbstring_big5string(commit_str,for_big5,strlen(commit_str));
      strcpy(commit_str,for_big5);
      setlocale(LC_CTYPE,BIG5_LC_CTYPE);
    }else{
      setlocale(LC_CTYPE,GB_LC_CTYPE);
    }
    XmbTextListToTextProperty(ims->core.display, (char **) &text, 1,
			      XCompoundTextStyle, &tp);
    ((IMCommitStruct *) call_data)->flag |= XimLookupChars;
    ((IMCommitStruct *) call_data)->commit_string = (char *)tp.value;
    IMCommitString(ims, (XPointer) call_data);
    XFree(tp.value);
    commit_str[0]='\0';
    return True;
  case NEED_BYPASS:
    return True;
  case NEED_FORWARD:
    IMForwardEvent(ims, (XPointer) call_data);
    return True;
  default:
    return False;		/* should never happen */
  }
  /* In case of Static Event Flow */
  /*
    if (!use_trigger) {
    static Bool     preedit_state_flag = False;
    if (IsMatchKeys(ims, call_data, Trigger_Keys)) {
    preedit_state_flag = !preedit_state_flag;
    return True;
    }
    }
  */

}


Bool
im_trigger_notify(register XIMS ims, register IMTriggerNotifyStruct * call_data)
{
	if (call_data->flag == 0) {	/* on key */
	  XMapRaised(display, w_input);
	  IMPreeditStart(ims, (XPointer) call_data);
		/*
		 * Here, the start of preediting is notified from IMlibrary,
		 * which is the only way to start preediting in case of
		 * Dynamic Event Flow, because ON key is mandatary for
		 * Dynamic Event Flow.
		 */
		return True;
	} else if (call_data->flag == 1) {	/* off key */
	  XUnmapWindow(display, w_input);
	  IMPreeditEnd(ims, (XPointer) call_data);
		/*
		 * Here, the end of preediting is notified from the
		 * IMlibrary, which happens only if OFF key, which is
		 * optional for Dynamic Event Flow, has been registered by
		 * IMOpenIM or IMSetIMValues, otherwise, the end of
		 * preediting must be notified from the IMserver to the
		 * IMlibrary.
		 */
		return True;
	} else {
		/* never happens */
		return False;
	}
}

Bool
im_preedit_start_reply(register XIMS ims, register IMPreeditCBStruct * call_data)
{
	return True;
}

Bool
im_preedit_caret_reply(register XIMS ims, register IMPreeditCBStruct * call_data)
{
	return True;
}

Bool
im_protocol_handler(register XIMS ims, register IMProtocol * call_data)
{
	switch (call_data->major_code) {
	case XIM_OPEN:
		return im_open(ims, (IMOpenStruct *) call_data);
	case XIM_CLOSE:
		return im_close(ims, (IMCloseStruct *) call_data);
	case XIM_CREATE_IC:
		return im_create_ic(ims, (IMChangeICStruct *) call_data);
	case XIM_DESTROY_IC:
		return im_destroy_ic(ims, (IMChangeICStruct *) call_data);
	case XIM_SET_IC_VALUES:
		return im_set_ic_value(ims, (IMChangeICStruct *) call_data);
	case XIM_GET_IC_VALUES:
		return im_get_ic_value(ims, (IMChangeICStruct *) call_data);
	case XIM_FORWARD_EVENT:
		return im_forward_event(ims, (IMForwardEventStruct *) call_data);
	case XIM_SET_IC_FOCUS:
		return im_set_ic_focus(ims, (IMChangeFocusStruct *) call_data);
	case XIM_UNSET_IC_FOCUS:
		return im_unset_ic_focus(ims, (IMChangeFocusStruct *) call_data);
	case XIM_RESET_IC:
		return im_reset_ic(ims, (IMResetICStruct *) call_data);
	case XIM_TRIGGER_NOTIFY:
		return im_trigger_notify(ims, (IMTriggerNotifyStruct *) call_data);
	case XIM_PREEDIT_START_REPLY:
		return im_preedit_start_reply(ims, (IMPreeditCBStruct *) call_data);
	case XIM_PREEDIT_CARET_REPLY:
		return im_preedit_caret_reply(ims, (IMPreeditCBStruct *) call_data);
	default:
		break;
	}
	return True;
}

void
set_xim(void)
{

	if ((input_styles = (XIMStyles *) malloc(sizeof(XIMStyles))) == NULL)
		err(ERR_NO, "%s: can not allocate memory for XIMStyles\n", progname);

	input_styles->count_styles = sizeof(Styles) / sizeof(XIMStyle) - 1;
	input_styles->supported_styles = Styles;

	if ((encodings = (XIMEncodings *)malloc(sizeof(XIMEncodings)))==NULL) {
	  err(ERR_NO,"Can't allocate\n");
    	}
    	encodings->count_encodings = sizeof(chEncodings)/sizeof(XIMEncoding)-1;
    	encodings->supported_encodings = chEncodings;

	if ((on_keys = (XIMTriggerKeys *) malloc(sizeof(XIMTriggerKeys))) == NULL)
		err(ERR_NO, "%s: cannot allocate memory for XIMTriggerKeys\n", progname);

	on_keys->count_keys = sizeof(Trigger_Keys) / sizeof(XIMTriggerKey) - 1;
	on_keys->keylist = Trigger_Keys;

	/* only use X Input Method Protocol */
	ims = IMOpenIM(display,
		       IMModifiers, "Xi18n",
		       IMServerWindow, w_input,
		       IMServerName, DEFAULT_IM_NAME,
		       IMLocale, DEFAULT_LOCALE,
		       IMServerTransport, DEFAULT_TRANSPORT,
		       IMInputStyles, input_styles,
		       IMEncodingList, 	encodings,
		       NULL);
	if (ims == (XIMS) NULL)
	  err(ERR_NO, "Can't Open Input Method %s:%s\n", DEFAULT_IM_NAME, 
	      DEFAULT_TRANSPORT);

	IMSetIMValues(ims,
		      IMOnKeysList, on_keys,
		      IMOffKeysList, on_keys,
		      NULL);
	IMSetIMValues(ims,
		      IMProtocolHandler, im_protocol_handler,
		      IMFilterEventMask, KeyPressMask,
		      NULL);
       
}
