%{
/*-------------------------------------------------------------------
	b5addpy.l

	Guo Jin (guojin@iss.nus.sg)     Oct. 29, 1994

	To add pinyin annotation to certain BIG5 hanzi:
	o rare hanzi (frequency less than a setted threshold);
	o randomly selected (to help non-native simplified hanzi reader).

	usage: b5addpy [-h] [-f frequency] [-r probability]
		-h:	for help.
		-f:	the frequency threshold (annotate only those with
			frequency no more than the threshold).
		-r:	the probability threshold (randomly annotate the
			setted probability of hanz).
	default:
		FREQ = 5;
		PROB = 0.01;
-------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>

#define	FREQ_LEFT	"i"
#define	FREQ_RIGHT	"j"
#define	PROB_LEFT	"e"
#define	PROB_RIGHT	"f"

long	FREQ = 5;	/* default maximum emerging frequency */
double	PROB = 0.01;	/* annotate PROB number of all hanzi */
unsigned short xsubi[3];

int b5_proc(char *yytext);
%}

BIG5	([\xa1-\xc6]|[\xc9-\xf9])([\x40-\x7e]|[\xa1-\xfe])

%%

{BIG5}	{ b5_proc(yytext); }

%%

#include <math.h>

typedef struct {
        char    hanzi[3];
        long    freq;
        char    pinyin[8];
} HANZI;

#include "b5addpy.dat"

int b5_proc(char *yytext)
{
        int     index;
        unsigned char   ch1, ch2, qu, wei;
	double	prob;

        ch1 = yytext[0];
        ch2 = yytext[1];
	qu = (ch1 <= 0xc6) ? (ch1-0xa1) : (ch1-0xa1-0x02);
	wei = (ch2 <= 0x7e) ? (ch2-0x40) : (ch2-0x40-0x22);

        index = qu*157 + wei;

	/* printf("%2x%2x  index=%d, <%s>", ch1, ch2, index, yytext); */
 
	printf("%s", b5_hanzi[index].hanzi);
	if (strlen(b5_hanzi[index].pinyin) != 0) {
		if (b5_hanzi[index].freq <= FREQ) {
			printf("%s%s%s",
				FREQ_LEFT, b5_hanzi[index].pinyin, FREQ_RIGHT);
		} else {
			prob = (double)lrand48()/INT_MAX;
			/* printf("[PROB=%f, prob=%f]\n", PROB, prob); */
			if (prob <= PROB) {
			printf("%s%s%s",
				PROB_LEFT, b5_hanzi[index].pinyin, PROB_RIGHT);
			}
		}
	}
}

command_help()
{
	printf("usage: b5addpy [-h] [-f frequency] [-r probability]\n");
	printf("default: frequency=%d, probability=%f\n", FREQ, PROB);
	exit(0);
}

main(int argc, char *argv[])
{
	int	k;

	k=1;
	while(k<argc) {
		if (argv[k][0] == '-') {
			switch(argv[k][1]) {
			case 'f':
				FREQ = atoi(argv[k+1]);
				k += 2;
				break;
			case 'r':
				sscanf(argv[k+1], "%lf", &PROB);
				k += 2;
				break;
			case 'h':
			default:
				command_help();
			}
		} else {
			k++;
		}
	}

	yylex();

	return 0;
}
