%{
/*-------------------------------------------------------------------
	gbaddpy.l

	Guo Jin (guojin@iss.nus.sg)     Oct. 29, 1994

	To add pinyin annotation to certain GB hanzi:
	o rare hanzi (frequency less than a setted threshold);
	o randomly selected (to help non-native simplified hanzi reader).

	usage: gbaddpy [-h] [-f frequency] [-r probability]
		-h:	for help.
		-f:	the frequency threshold (annotate only those with
			frequency no more than the threshold).
		-r:	the probability threshold (randomly annotate the
			setted probability of hanz).
	default:
		FREQ = 5;
		PROB = 0.01;
-------------------------------------------------------------------*/
#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <math.h>

#define FREQ_LEFT	""
#define FREQ_RIGHT	""
#define PROB_LEFT	""
#define PROB_RIGHT	""

long	FREQ = 5;	/* default maximum emerging frequency */
double	PROB = 0.01;	/* annotate PROB number of all hanzi */
unsigned short xsubi[3];

int gb_proc(char *yytext);
%}

GB	([\xa1-\xf7][\xa1-\xfe])

%%

{GB}	{ gb_proc(yytext); }

%%

typedef struct {
        char    hanzi[3];
        long    freq;
        char    pinyin[8];
} HANZI;

#include "gbaddpy.dat"

int gb_proc(char *yytext)
{
        int     index;
        unsigned char   ch1, ch2;
	double	prob;

        ch1 = yytext[0];
        ch2 = yytext[1];
        index = (ch1-0xa1)*94 + (ch2-0xa1);

	/* printf("%2x%2x  index=%d, <%s>", ch1, ch2, index, yytext); */
 
	printf("%s", gb_hanzi[index].hanzi);
	if (strlen(gb_hanzi[index].pinyin) != 0) {
		if (gb_hanzi[index].freq <= FREQ) {
			 printf("%s%s%s",
				 FREQ_LEFT, gb_hanzi[index].pinyin, FREQ_RIGHT);
		} else {
			prob = (double)lrand48()/INT_MAX;
			/* printf("[PROB=%f, prob=%f]\n", PROB, prob); */
			if (prob <= PROB) {
			printf("%s%s%s",
				PROB_LEFT, gb_hanzi[index].pinyin, PROB_RIGHT);
			}
		}
	}
}

command_help()
{
	printf("usage: b5addpy [-h] [-f frequency] [-r probability]\n");
	printf("default: frequency=%d, probability=%f\n", FREQ, PROB);
	exit(0);
}

main(int argc, char *argv[])
{
	int	k;

	k=1;
	while(k<argc) {
		if (argv[k][0] == '-') {
			switch(argv[k][1]) {
			case 'f':
				FREQ = atoi(argv[k+1]);
				k += 2;
				break;
			case 'r':
				sscanf(argv[k+1], "%lf", &PROB);
				/* printf("argv=<%s>, PROB=%f\n", argv[k+1], PROB); */
				k += 2;
				break;
			case 'h':
			default:
				command_help();
			}
		} else {
			k++;
		}
	}

	yylex();

	return 0;
}
