#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <sys/param.h>
#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <mysql/mysql.h>
#include "db_util.h"
#include "utils.h"




 char    hostname[MAX_CONFIG_VARS_LEN];
 char    username[MAX_CONFIG_VARS_LEN];
 char    password[MAX_CONFIG_VARS_LEN];
 char    database[MAX_CONFIG_VARS_LEN];
 char    pydb_table[MAX_CONFIG_VARS_LEN];
 char    port[MAX_CONFIG_VARS_LEN];


/* 
 * compile mysqlserver with flag --with-charset=gb2312 to 
 * let mysql support chinese characters as primary key.
 *
  create database pydb;
  use pydb;
  create table pydb(
      pinyin varchar(255) not null,
      hanzi varchar(255) not null,
      freq float(23,0) ,
      last_used TIMESTAMP(14),
      primary key (hanzi),
      key(pinyin));
 */


extern MYSQL pydb;
extern HANZI **py;


void
add_all_gb_hz(MYSQL *db)
{
  int ret;
  char buf[]="ALL_GB_HANZI_INCLUDED";

  ret= is_have_one(db,buf);
  if( ret )
    init_py_ref_db(db,0);
  else{
    pydb_insert_direct(db,buf,buf,1);
    init_py_ref_db(db,1);
  }

}


void
gen_items(int argc, char **argv)
{
  u_char_8 *p,*p_head,*p_end;
  int len;
  int fd;
  int flag;

  add_all_gb_hz(&pydb);
  while (--argc) {
    flag = 0;
    p_head = p = (u_char_8 *) load_file(argv[argc], &len, &fd);
    if (p != NULL) {
      fprintf(stderr, "[%d]\t[%s]\n", len, argv[argc]);
      if(len < MAX_HANZI_CHAR_NUMBER*2){
	fprintf(stderr,"File size should be larger than %d.\n",MAX_HANZI_CHAR_NUMBER*2);
	continue;
      }
      for (p_end = p + len - MAX_HANZI_CHAR_NUMBER*2; p < p_end; p += flag) {
	flag=add_item(p,LOW_PRIORITY);
      }
    }
    if (p != NULL)
      unload_file(p_head, len, fd);
  }
}

#if 0
void
gen_table(void)
{
	register int    i, j;

	for (i = 160; i < 256; i++)
		for (j = 160; j < 256; j++)
			printf("(%x,%x)=[%c%c]\n", i, j, i, j);
	exit(0);
}
#endif



int
read_conf(char *config_file_path)
{
  FILE *in;
	in=fopen(config_file_path,"rb");
	if( !in ){
		perror("fopen()");
		return -1;
	}
#define IF_OK(x,y)	do{ if(!get_tag_content(in,x,y)){ \
				warn("read_conf(%s) error!\n",x); \
				fclose(in); \
				return -2; \
				}  \
			}while(0)
	IF_OK("hostname",hostname);
	IF_OK("username",username);
	IF_OK("password",password);
	IF_OK("database",database);
	IF_OK("table",pydb_table);
	IF_OK("port",port);

	fclose(in);
	return 0;
}



int
main(int argc,char *argv[])
{
  int ret=0;

  ret=read_conf(DEFAULT_CONF_PATH);
  if( ret < 0){
    fprintf(stderr,"Error in reading configuration file %s\n",DEFAULT_CONF_PATH);
    exit(-1);
  }
  (void)db_connect( &pydb,hostname,username,password,database,
		    atoi(port),NULL,0);
  gen_items(argc,argv);
  db_close(&pydb);
  free(py);
  exit(0);
}








